# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import re
import sys

# ==============================================================================
# CRITICAL CONFIGURATION
# ==============================================================================

# User agent for all requests, if not overwritten by Flaresolverr
FALLBACK_USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/145.0.0.0 Safari/537.36"

# Standard request timeout budgets.
# Slow mode multiplies each base timeout by this factor.
TIMEOUT_SLOW_MODE_MULTIPLIER = 3

# Table storing per-timeout slow mode flags.
TIMEOUT_SLOW_MODE_TABLE = "timeout_slow_mode"

# Timeout settings shown in the Web UI.
TIMEOUT_SLOW_MODE_DEFINITIONS = {
    "search": {"label": "Search Timeout", "base_seconds": 15},
    "feed": {"label": "Feed Timeout", "base_seconds": 30},
    "download": {"label": "Download Timeout", "base_seconds": 30},
    "session": {"label": "Session Timeout", "base_seconds": 30},
}


def _coerce_timeout_bool(value, default=False):
    if isinstance(value, bool):
        return value
    if value is None:
        return default
    if isinstance(value, str):
        lowered = value.strip().lower()
        if lowered in {"1", "true", "yes", "on"}:
            return True
        if lowered in {"0", "false", "no", "off"}:
            return False
    return bool(value)


def _is_timeout_slow_mode_enabled(timeout_key):
    try:
        from quasarr.providers import shared_state

        settings = shared_state.values.get("timeout_slow_mode")
        if isinstance(settings, dict):
            return _coerce_timeout_bool(settings.get(timeout_key), default=False)
    except Exception:
        return False

    return False


_TIMEOUT_RUNTIME_KEY_TO_NAME = {
    "search": "SEARCH_REQUEST_TIMEOUT_SECONDS",
    "feed": "FEED_REQUEST_TIMEOUT_SECONDS",
    "download": "DOWNLOAD_REQUEST_TIMEOUT_SECONDS",
    "session": "SESSION_REQUEST_TIMEOUT_SECONDS",
}


def _calculate_timeout_value(timeout_key, slow_mode_enabled):
    base_seconds = int(TIMEOUT_SLOW_MODE_DEFINITIONS[timeout_key]["base_seconds"])
    if slow_mode_enabled:
        return int(base_seconds * TIMEOUT_SLOW_MODE_MULTIPLIER)
    return base_seconds


def apply_timeout_slow_mode_settings(settings=None):
    if not isinstance(settings, dict):
        settings = {
            timeout_key: _is_timeout_slow_mode_enabled(timeout_key)
            for timeout_key in TIMEOUT_SLOW_MODE_DEFINITIONS
        }

    resolved_timeout_values = {}
    for timeout_key, const_name in _TIMEOUT_RUNTIME_KEY_TO_NAME.items():
        slow_mode_enabled = _coerce_timeout_bool(
            settings.get(timeout_key),
            default=False,
        )
        resolved_timeout_values[const_name] = _calculate_timeout_value(
            timeout_key,
            slow_mode_enabled,
        )

    for const_name, timeout_value in resolved_timeout_values.items():
        globals()[const_name] = timeout_value

    for module_name, module in list(sys.modules.items()):
        if module_name != "quasarr" and not module_name.startswith("quasarr."):
            continue

        module_dict = getattr(module, "__dict__", None)
        if not isinstance(module_dict, dict):
            continue

        for const_name, timeout_value in resolved_timeout_values.items():
            if const_name in module_dict:
                module_dict[const_name] = timeout_value

    return resolved_timeout_values


SEARCH_REQUEST_TIMEOUT_SECONDS = int(
    TIMEOUT_SLOW_MODE_DEFINITIONS["search"]["base_seconds"]
)
FEED_REQUEST_TIMEOUT_SECONDS = int(
    TIMEOUT_SLOW_MODE_DEFINITIONS["feed"]["base_seconds"]
)
DOWNLOAD_REQUEST_TIMEOUT_SECONDS = int(
    TIMEOUT_SLOW_MODE_DEFINITIONS["download"]["base_seconds"]
)
SESSION_REQUEST_TIMEOUT_SECONDS = int(
    TIMEOUT_SLOW_MODE_DEFINITIONS["session"]["base_seconds"]
)

# Notification providers exposed in config/UI.
NOTIFICATION_PROVIDERS = ("discord", "telegram")


# ==============================================================================
# SEARCH AND DOWNLOAD CATEGORIES
# ==============================================================================

# Search category source of truth
SEARCH_CATEGORY_DEFINITIONS = {
    "MOVIES": {"id": 2000, "name": "Movies", "emoji": "🎬"},
    "MOVIES_HD": {"id": 2040, "name": "Movies/HD", "emoji": "🎬"},
    "MOVIES_UHD": {"id": 2045, "name": "Movies/UHD", "emoji": "🎬"},
    "MUSIC": {"id": 3000, "name": "Audio", "emoji": "🎵"},
    "MUSIC_MP3": {"id": 3010, "name": "Audio/MP3", "emoji": "🎵"},
    "MUSIC_FLAC": {"id": 3040, "name": "Audio/FLAC", "emoji": "🎵"},
    "SHOWS": {"id": 5000, "name": "TV", "emoji": "📺"},
    "SHOWS_HD": {"id": 5040, "name": "TV/HD", "emoji": "📺"},
    "SHOWS_UHD": {"id": 5045, "name": "TV/UHD", "emoji": "📺"},
    "SHOWS_ANIME": {"id": 5070, "name": "TV/Anime", "emoji": "⛩️"},
    "SHOWS_DOCUMENTARY": {"id": 5080, "name": "TV/Documentary", "emoji": "🎥"},
    "XXX": {"id": 6000, "name": "XXX", "emoji": "🔞"},
    "BOOKS": {"id": 7000, "name": "Books", "emoji": "📚"},
}

# Importable constants for search categories are generated from the source of truth above to ensure consistency and maintainability.
SEARCH_CAT_MOVIES = SEARCH_CATEGORY_DEFINITIONS["MOVIES"]["id"]
SEARCH_CAT_MOVIES_HD = SEARCH_CATEGORY_DEFINITIONS["MOVIES_HD"]["id"]
SEARCH_CAT_MOVIES_UHD = SEARCH_CATEGORY_DEFINITIONS["MOVIES_UHD"]["id"]
SEARCH_CAT_MUSIC = SEARCH_CATEGORY_DEFINITIONS["MUSIC"]["id"]
SEARCH_CAT_MUSIC_MP3 = SEARCH_CATEGORY_DEFINITIONS["MUSIC_MP3"]["id"]
SEARCH_CAT_MUSIC_FLAC = SEARCH_CATEGORY_DEFINITIONS["MUSIC_FLAC"]["id"]
SEARCH_CAT_SHOWS = SEARCH_CATEGORY_DEFINITIONS["SHOWS"]["id"]
SEARCH_CAT_SHOWS_HD = SEARCH_CATEGORY_DEFINITIONS["SHOWS_HD"]["id"]
SEARCH_CAT_SHOWS_UHD = SEARCH_CATEGORY_DEFINITIONS["SHOWS_UHD"]["id"]
SEARCH_CAT_SHOWS_ANIME = SEARCH_CATEGORY_DEFINITIONS["SHOWS_ANIME"]["id"]
SEARCH_CAT_SHOWS_DOCUMENTARY = SEARCH_CATEGORY_DEFINITIONS["SHOWS_DOCUMENTARY"]["id"]
SEARCH_CAT_XXX = SEARCH_CATEGORY_DEFINITIONS["XXX"]["id"]
SEARCH_CAT_BOOKS = SEARCH_CATEGORY_DEFINITIONS["BOOKS"]["id"]

SEARCH_CATEGORIES = {
    definition["id"]: {"name": definition["name"], "emoji": definition["emoji"]}
    for definition in SEARCH_CATEGORY_DEFINITIONS.values()
}

# Explicit cache-sharing families for search result population.
# The owner key represents the category that should populate cache first whenever possible.
# Categories not listed here are treated as standalone cache groups.
SEARCH_CATEGORY_CACHE_FAMILIES = {
    SEARCH_CAT_MOVIES: (
        SEARCH_CAT_MOVIES,
        SEARCH_CAT_MOVIES_HD,
        SEARCH_CAT_MOVIES_UHD,
    ),
    SEARCH_CAT_MUSIC: (
        SEARCH_CAT_MUSIC,
        SEARCH_CAT_MUSIC_MP3,
        SEARCH_CAT_MUSIC_FLAC,
    ),
    SEARCH_CAT_SHOWS: (
        SEARCH_CAT_SHOWS,
        SEARCH_CAT_SHOWS_HD,
        SEARCH_CAT_SHOWS_UHD,
    ),
    SEARCH_CAT_SHOWS_ANIME: (SEARCH_CAT_SHOWS_ANIME,),
    SEARCH_CAT_SHOWS_DOCUMENTARY: (SEARCH_CAT_SHOWS_DOCUMENTARY,),
    SEARCH_CAT_XXX: (SEARCH_CAT_XXX,),
    SEARCH_CAT_BOOKS: (SEARCH_CAT_BOOKS,),
}

# Default Set of Download Categories
DOWNLOAD_CATEGORIES = {
    "movies": {"emoji": "🎬"},
    "music": {"emoji": "🎵"},
    "tv": {"emoji": "📺"},
    "docs": {"emoji": "📄"},
}

# Fallback category mapping by client type (used when no valid download category is provided)
CLIENT_DOWNLOAD_CATEGORY_FALLBACK_MAP = {
    "lazylibrarian": "docs",
    "lidarr": "music",
    "radarr": "movies",
    "sonarr": "tv",
}

# ==============================================================================
# HOSTERS & MIRRORS
# ==============================================================================

# As of 01.02.2026
# Format: (Name, is_tier1)
HOSTERS = [
    # --- TIER 1: The Standards (Required for most downloaders) ---
    ("Rapidgator", True),  # Global King. Most files are here.
    ("DDownload", True),  # The "Euro Standard". Cheaper alternative to RG.
    # --- TIER 2: Very Popular / High Retention ---
    ("1fichier", False),  # Massive retention, cheap, very popular in France/Global.
    ("Keep2Share", False),  # "Premium" tier. High speeds, expensive, very stable.
    (
        "Nitroflare",
        False,
    ),  # Old guard. Expensive, but essential for some exclusive content.
    # --- TIER 3: Common Mirrors (The "Third Link") ---
    ("Turbobit", False),  # Everywhere, but often disliked by free users.
    ("Hitfile", False),  # Turbobit's sibling site. Often seen together.
    ("Katfile", False),  # Very common secondary mirror for smaller uploaders.
    ("Alfafile", False),  # Stable mid-tier host, often seen on DDL blogs.
    # --- TIER 4: Niche / Backup / User Requested ---
    ("Filer", False),  # Strong in German-speaking areas, niche elsewhere.
    (
        "IronFiles",
        False,
    ),  # Active. Smaller ecosystem, often specific to certain boards.
    ("Fikper", False),  # Newer player (relative to RG), gained traction in 2024-25.
    ("Mega", False),  # Active, but functions differently (cloud drive vs. OCH).
    ("AkiraBox", False),  # Common mirror on anime-focused releases.
    ("BuzzHeavier", False),  # Common mirror on anime-focused releases.
    ("ClickNupload", False),  # Common mirror on RM releases.
    ("GoFile", False),  # Common mirror on anime-focused releases.
    ("KrakenFiles", False),  # Common mirror on anime-focused releases.
    ("MdiaLoad", False),  # Common mirror on anime-focused releases.
    ("MultiUp", False),  # Common mirror on anime-focused releases.
    ("RapidRAR", False),  # Common mirror on RM releases.
    ("frdl", False),  # Rarely used in some forums
]

# Used to identify share hosters (lowercase for comparison)
SHARE_HOSTERS_LOWERCASE = {h[0].lower() for h in HOSTERS}
SHARE_HOSTERS = [h[0] for h in HOSTERS]

# Recommend only these
RECOMMENDED_HOSTERS = [h[0] for h in HOSTERS if h[1]]

# Evidence-backed root-token aliases for final mirror whitelist checks.
# Sources used on 2026-03-02:
# - JDownloader host plugins (mirror/jdownloader)
# - pyLoad host plugins (pyload/pyload)
#
# Matching is performed on the pure mirror/root token derived from the final
# HTTP hostname, ignoring TLDs and subdomains. Keep only root tokens here.
MIRROR_TOKEN_ALIASES = {
    "alterupload": "1fichier",
    "cjoint": "1fichier",
    "clickndownload": "clicknupload",
    "ddl": "ddownload",
    "ddlto": "ddownload",
    "depositfiles": "turbobit",
    "desfichiers": "1fichier",
    "dfichiers": "1fichier",
    "dl4free": "1fichier",
    "dlbit": "turbobit",
    "fayloobmennik": "turbobit",
    "fboom": "keep2share",
    "filedeluxe": "turbobit",
    "fileboom": "keep2share",
    "filemaster": "turbobit",
    "flacmania": "turbobit",
    "filernet": "filer",
    "filhost": "turbobit",
    "hil": "hitfile",
    "hotshare": "turbobit",
    "ifolder": "turbobit",
    "k2s": "keep2share",
    "k2share": "keep2share",
    "keep2": "keep2share",
    "keep2s": "keep2share",
    "megadl": "1fichier",
    "mesfichiers": "1fichier",
    "nitro": "nitroflare",
    "pjointe": "1fichier",
    "piecejointe": "1fichier",
    "publish2": "keep2share",
    "rapidfile": "turbobit",
    "rg": "rapidgator",
    "sibit": "turbobit",
    "tb": "turbobit",
    "tenvoi": "1fichier",
    "tezfiles": "keep2share",
    "tourbobit": "turbobit",
    "trbbt": "turbobit",
    "trubobit": "turbobit",
    "turb": "turbobit",
    "turbo": "turbobit",
    "turbabit": "turbobit",
    "turbobeet": "turbobit",
    "turbobi": "turbobit",
    "turbobbit": "turbobit",
    "turbobif": "turbobit",
    "turbobit5": "turbobit",
    "turbobita": "turbobit",
    "turbobite": "turbobit",
    "turbobith": "turbobit",
    "turbobitn": "turbobit",
    "turbobitt": "turbobit",
    "turbobiyt": "turbobit",
    "turbobyt": "turbobit",
    "turbobyte": "turbobit",
    "turboot": "turbobit",
    "turboobit": "turbobit",
    "turobit": "turbobit",
    "twobit": "turbobit",
    "ucdn": "ddownload",
    "wayupload": "turbobit",
    "xrfiles": "turbobit",
}

# ==============================================================================
# REGEX PATTERNS (CONTENT & PARSING)
# ==============================================================================

# Used to identify release tittles with season or Episode numbers
SEASON_EP_REGEX = re.compile(
    r"(?i)(?:S\d{1,3}(?:E\d{1,3}(?:-\d{1,3})?)?|S\d{1,3}-\d{1,3})"
)

# Used to identify movies
MOVIE_REGEX = re.compile(
    r"^(?!.*(?:S\d{1,3}(?:E\d{1,3}(?:-\d{1,3})?)?|S\d{1,3}-\d{1,3})).*$", re.IGNORECASE
)

# Domain checks
AFFILIATE_REGEX = re.compile(r"af\.php\?v=([a-zA-Z0-9]+)")
FILECRYPT_REGEX = re.compile(
    r"https?://(?:www\.)?filecrypt\.(?:cc|co|to)/[Cc]ontainer/[A-Za-z0-9]+", re.I
)
IMDB_REGEX = re.compile(r"imdb\.com/title/(tt\d+)", re.I)

# Release Title Checks
RESOLUTION_REGEX = re.compile(r"\d{3,4}p", re.I)
CODEC_REGEX = re.compile(r"x264|x265|h264|h265|hevc|avc", re.I)
XXX_REGEX = re.compile(r"\.xxx\.", re.I)

SIZE_REGEX = re.compile(r"Größe[:\s]*(\d+(?:[.,]\d+)?)\s*(MB|GB|TB)", re.I)

DATE_REGEX = re.compile(r"(\d{2}\.\d{2}\.\d{2}),?\s*(\d{1,2}:\d{2})")

# Pattern to extract individual episode release names from text
# Matches: Title.S02E03.Info-GROUP (group name starts after hyphen)
EPISODE_EXTRACT_REGEX = re.compile(
    r"([A-Za-z][A-Za-z0-9.]+\.S\d{2}E\d{2}[A-Za-z0-9.]*-[A-Za-z][A-Za-z0-9]*)", re.I
)

# Pattern to clean trailing common words that may be attached to group names
# e.g., -WAYNEAvg -> -WAYNE, -GROUPBitrate -> -GROUP
TRAILING_GARBAGE_PATTERN = re.compile(
    r"(Avg|Bitrate|Size|Größe|Video|Audio|Duration|Release|Info).*$", re.I
)

# Pattern to extract average bitrate (e.g., "Avg. Bitrate: 10,6 Mb/s" or "6 040 kb/s")
# Note: Numbers may contain spaces as thousand separators (e.g., "6 040")
BITRATE_REGEX = re.compile(
    r"(?:Avg\.?\s*)?Bitrate[:\s]*([\d\s]+(?:[.,]\d+)?)\s*(kb/s|Mb/s|mb/s)", re.I
)

# Pattern to extract episode duration (e.g., "Dauer: 60 Min. pro Folge")
EPISODE_DURATION_REGEX = re.compile(r"Dauer[:\s]*(\d+)\s*Min\.?\s*pro\s*Folge", re.I)


# ==============================================================================
# DECRYPTION & PROTECTION
# ==============================================================================

# Identifies Linkcrypters that never require CAPTCHA
AUTO_DECRYPT_PATTERNS = {
    "hide": re.compile(r"hide\.", re.IGNORECASE),
}

# Identifies Linkcrypters that may require CAPTCHA
PROTECTED_PATTERNS = {
    "filecrypt": re.compile(r"filecrypt\.", re.IGNORECASE),
    "tolink": re.compile(r"tolink\.", re.IGNORECASE),
    "keeplinks": re.compile(r"keeplinks\.", re.IGNORECASE),
}


# ==============================================================================
# QUASARR PACKAGE MANAGEMENT
# ==============================================================================

# Prefix for all Quasarr Packages
PACKAGE_ID_PREFIX = "Quasarr_"

# Regex for strict Quasarr ID validation: Quasarr_{category}_{32_char_hash}
PACKAGE_ID_PATTERN = re.compile(r"^Quasarr_[a-z]+_[a-f0-9]{32}$")


# ==============================================================================
# JDOWNLOADER / EXTRACTION
# ==============================================================================

# Status Strings from JDownloader (add more languages if needed)
EXTRACTION_COMPLETE_MARKERS = (
    "extraction ok",  # English
    "entpacken ok",  # German
)

# Used during archive detection
ARCHIVE_EXTENSIONS = frozenset(
    [
        ".rar",
        ".zip",
        ".7z",
        ".tar",
        ".gz",
        ".bz2",
        ".xz",
        ".001",
        ".002",
        ".003",
        ".004",
        ".005",
        ".006",
        ".007",
        ".008",
        ".009",
        ".r00",
        ".r01",
        ".r02",
        ".r03",
        ".r04",
        ".r05",
        ".r06",
        ".r07",
        ".r08",
        ".r09",
        ".part1.rar",
        ".part01.rar",
        ".part001.rar",
        ".part2.rar",
        ".part02.rar",
        ".part002.rar",
    ]
)


# ==============================================================================
# TIME & LOCALIZATION
# ==============================================================================

LANGUAGE_TO_ALPHA2 = {
    "german": "DE",
    "deutsch": "DE",
    "de": "DE",
    "ger": "DE",
    "deu": "DE",
    "english": "EN",
    "englisch": "EN",
    "en": "EN",
    "eng": "EN",
    "japanese": "JP",
    "japanisch": "JP",
    "jp": "JP",
    "jpn": "JP",
}

SUBTITLE_TOKEN_BY_ALPHA2 = {
    "DE": "GerSub",
    "EN": "EngSub",
    "JP": "JapSub",
}

SESSION_MAX_AGE_SECONDS = 24 * 60 * 60  # 24 hours

# Source of truth for month names and mappings
# Format: (Month Number, English Name, German Name, [Additional Synonyms])
_MONTHS_CONFIG = [
    (1, "January", "Januar", ["jan"]),
    (2, "February", "Februar", ["feb"]),
    (3, "March", "März", ["maerz", "mär", "mrz", "mae"]),
    (4, "April", "April", ["apr"]),
    (5, "May", "Mai", ["may"]),
    (6, "June", "Juni", ["jun", "june"]),
    (7, "July", "Juli", ["jul", "july"]),
    (8, "August", "August", ["aug"]),
    (9, "September", "September", ["sep"]),
    (10, "October", "Oktober", ["okt", "october"]),
    (11, "November", "November", ["nov"]),
    (12, "December", "Dezember", ["dez", "december"]),
]

ENGLISH_MONTHS = [m[1] for m in _MONTHS_CONFIG]
GERMAN_MONTHS = [m[2] for m in _MONTHS_CONFIG]

MONTHS_MAP = {
    name.lower(): nr
    for nr, en, de, synonyms in _MONTHS_CONFIG
    for name in [en, de] + synonyms
}


# ==============================================================================
# DISCORD / NOTIFICATIONS
# ==============================================================================

# Shared notification branding/assets
QUASARR_AVATAR = "https://raw.githubusercontent.com/rix1337/Quasarr/main/Quasarr.png"
SPONSORS_HELPER_URL = (
    "https://github.com/rix1337/Quasarr?tab=readme-ov-file#sponsorshelper"
)

# Discord message flag for suppressing notifications
SUPPRESS_NOTIFICATIONS = 1 << 12  # 4096



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

from quasarr.storage.setup.common import (
    add_no_cache_headers,
    render_reconnect_success,
    setup_auth,
)
from quasarr.storage.setup.flaresolverr import (
    delete_skip_flaresolverr_preference,
    flaresolverr_config,
    flaresolverr_form_html,
    get_flaresolverr_status_data,
    save_flaresolverr_url,
)
from quasarr.storage.setup.hostnames import (
    check_credentials,
    clear_skip_login,
    get_skip_login,
    hostname_credentials_config,
    hostname_form_html,
    hostnames_config,
    import_hostnames_from_url,
    save_hostnames,
)
from quasarr.storage.setup.jdownloader import (
    jdownloader_config,
    save_jdownloader_settings,
    verify_jdownloader_credentials,
)
from quasarr.storage.setup.notifications import (
    get_notification_settings_data,
    initialize_notification_settings,
    refresh_notification_settings,
    save_notification_settings,
    send_notification_test,
)
from quasarr.storage.setup.path import path_config
from quasarr.storage.setup.timeouts import (
    get_timeout_slow_mode_settings_data,
    initialize_timeout_slow_mode_settings,
    refresh_timeout_slow_mode_settings,
    save_timeout_slow_mode_settings,
)

__all__ = [
    "add_no_cache_headers",
    "check_credentials",
    "clear_skip_login",
    "delete_skip_flaresolverr_preference",
    "flaresolverr_config",
    "flaresolverr_form_html",
    "get_flaresolverr_status_data",
    "get_notification_settings_data",
    "get_skip_login",
    "hostname_credentials_config",
    "hostname_form_html",
    "hostnames_config",
    "import_hostnames_from_url",
    "initialize_notification_settings",
    "jdownloader_config",
    "path_config",
    "refresh_notification_settings",
    "render_reconnect_success",
    "save_flaresolverr_url",
    "save_hostnames",
    "save_jdownloader_settings",
    "save_notification_settings",
    "save_timeout_slow_mode_settings",
    "send_notification_test",
    "setup_auth",
    "get_timeout_slow_mode_settings_data",
    "initialize_timeout_slow_mode_settings",
    "refresh_timeout_slow_mode_settings",
    "verify_jdownloader_credentials",
]



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

from quasarr.providers.log import info
from quasarr.providers.notifications.helpers import resolve_poster_url
from quasarr.providers.notifications.helpers.message_builder import (
    build_notification_message,
)
from quasarr.providers.notifications.helpers.notification_types import (
    NotificationType,
    normalize_notification_type,
)


def _get_notification_settings(shared_state):
    settings = shared_state.values.get("notification_settings")
    return settings if isinstance(settings, dict) else {}


def _provider_case_enabled(shared_state, provider, notification_type):
    toggles = _get_notification_settings(shared_state).get("toggles")
    if not isinstance(toggles, dict):
        return True

    provider_toggles = toggles.get(provider)
    if not isinstance(provider_toggles, dict):
        return True

    return bool(provider_toggles.get(notification_type.value, True))


def _provider_case_silent(shared_state, provider, notification_type):
    silent_settings = _get_notification_settings(shared_state).get("silent")
    if not isinstance(silent_settings, dict):
        return False

    provider_silent = silent_settings.get(provider)
    if not isinstance(provider_silent, dict):
        return False

    return bool(provider_silent.get(notification_type.value, False))


def send_notification(
    shared_state, title, case, imdb_id=None, details=None, source=None
):
    """
    Send a notification to all configured providers (Discord, Telegram).

    Each provider is attempted independently — a failure in one does not block others.

    :param shared_state: Shared state object containing configuration.
    :param title: Title of the notification.
    :param case: A string representing the scenario (e.g., 'captcha', 'failed', 'unprotected').
    :param imdb_id: A string starting with "tt" followed by at least 7 digits, representing an object on IMDb
    :param details: A dictionary containing additional details, such as version and link for updates.
    :param source: Optional source of the notification, sent as a field in the embed.
    :return: True if at least one provider sent successfully, False otherwise.
    """
    from quasarr.providers.notifications import discord, telegram

    notification_type = normalize_notification_type(case)
    if notification_type is None:
        info(f"Unknown notification case: {case}")
        return False

    notification_settings = _get_notification_settings(shared_state)
    has_discord = bool(notification_settings.get("discord_webhook"))
    has_telegram = bool(notification_settings.get("telegram_bot_token")) and bool(
        notification_settings.get("telegram_chat_id")
    )

    if not has_discord and not has_telegram:
        return False

    # Resolve poster image once for all providers.
    image_url = None
    if notification_type in (NotificationType.UNPROTECTED, NotificationType.CAPTCHA):
        image_url = resolve_poster_url(shared_state, title, imdb_id)

    message = build_notification_message(
        shared_state,
        title,
        notification_type,
        details=details,
        source=source,
        image_url=image_url,
    )
    if message is None:
        return False

    any_success = False

    if has_discord and _provider_case_enabled(
        shared_state, "discord", notification_type
    ):
        discord_silent = _provider_case_silent(
            shared_state, "discord", notification_type
        )
        try:
            if discord.send(shared_state, message, silent=discord_silent):
                any_success = True
        except Exception as e:
            info(f"Discord notification error: {e}")

    if has_telegram and _provider_case_enabled(
        shared_state, "telegram", notification_type
    ):
        telegram_silent = _provider_case_silent(
            shared_state, "telegram", notification_type
        )
        try:
            if telegram.send(shared_state, message, silent=telegram_silent):
                any_success = True
        except Exception as e:
            info(f"Telegram notification error: {e}")

    return any_success



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

from urllib.parse import urljoin, urlparse

import requests
from bs4 import BeautifulSoup

from quasarr.constants import DOWNLOAD_REQUEST_TIMEOUT_SECONDS
from quasarr.downloads.sources.helpers.abstract_source import AbstractDownloadSource
from quasarr.providers.hostname_issues import mark_hostname_issue
from quasarr.providers.log import debug, info


class Source(AbstractDownloadSource):
    initials = "nk"

    def get_download_links(self, shared_state, url, mirrors, title, password):
        """
        NK source handler - fetches protected download links from NK pages.
        """
        requested_mirrors = {
            _normalize_mirror_name(mirror) for mirror in (mirrors or []) if mirror
        }

        host = shared_state.values["config"]("Hostnames").get(Source.initials)
        headers = {
            "User-Agent": shared_state.values["user_agent"],
        }

        session = requests.Session()

        try:
            r = session.get(
                url,
                headers=headers,
                timeout=DOWNLOAD_REQUEST_TIMEOUT_SECONDS,
            )
            r.raise_for_status()
            soup = BeautifulSoup(r.text, "html.parser")
        except Exception as e:
            info(f"Could not fetch release page for {title}: {e}")
            mark_hostname_issue(
                Source.initials,
                "download",
                str(e) if "e" in dir() else "Download error",
            )
            return {"links": []}

        anchors = soup.select("a.btn-orange")
        candidates = []
        for a in anchors:
            mirror = _normalize_mirror_name(a.text.strip())

            if mirror not in _SUPPORTED_MIRRORS:
                continue

            if requested_mirrors and mirror not in requested_mirrors:
                continue

            href = a.get("href", "").strip()
            if not href.lower().startswith(("http://", "https://")):
                href = "https://" + host + href

            if _is_nk_redirect_link(href, host):
                resolved_href = _resolve_nk_redirect(session, href, headers, title)
                if resolved_href:
                    href = resolved_href
                else:
                    # Never yield unresolved NK "/go" links. They bypass protected-link
                    # classification and end up as broken direct links in JD.
                    continue

            candidates.append([href, mirror])

        if not candidates:
            info(f"No external download links found for {title}")

        return {"links": candidates}


_SUPPORTED_MIRRORS = {"rapidgator", "ddownload"}


def _normalize_host(host):
    normalized = (host or "").lower().strip()
    if normalized.startswith("www."):
        normalized = normalized[4:]
    return normalized


def _is_nk_redirect_link(url, host):
    try:
        parsed = urlparse(url)
    except Exception:
        return False

    if _normalize_host(parsed.netloc) != _normalize_host(host):
        return False

    return parsed.path.startswith("/go/")


def _is_filecrypt_link(url):
    return "filecrypt." in (url or "").lower()


def _resolve_nk_redirect(session, url, headers, title):
    current_url = url
    visited = set()

    # Resolve redirect chain manually. This avoids fetching FileCrypt itself,
    # which can return 403 for automated traffic even when URL is valid.
    for _hop in range(8):
        if current_url in visited:
            debug(f"Could not resolve NK redirect for {title}: redirect loop detected")
            return None
        visited.add(current_url)

        try:
            response = session.get(
                current_url,
                headers=headers,
                allow_redirects=False,
                timeout=DOWNLOAD_REQUEST_TIMEOUT_SECONDS,
            )
        except requests.RequestException as e:
            debug(f"Could not resolve NK redirect for {title}: {e}")
            return None
        except Exception as e:
            debug(f"Could not resolve NK redirect for {title}: {e}")
            return None

        location = (response.headers.get("Location") or "").strip()
        if location:
            next_url = urljoin(current_url, location)
            if _is_filecrypt_link(next_url):
                return next_url
            current_url = next_url
            continue

        final_url = (response.url or current_url).strip()
        if _is_filecrypt_link(final_url):
            return final_url

        if "/404.html" in final_url:
            debug(f"NK redirect resolved to 404 for {title}: {final_url}")
            return None

        if response.status_code >= 400:
            debug(
                f"Could not resolve NK redirect for {title}: HTTP {response.status_code} at {final_url}"
            )
            return None

        debug(
            f"Could not resolve NK redirect for {title}: no redirect target from {final_url}"
        )
        return None

    debug(f"Could not resolve NK redirect for {title}: exceeded redirect hop limit")
    return None


def _normalize_mirror_name(mirror_name):
    normalized = mirror_name.lower().strip()

    if "://" in normalized:
        parsed = urlparse(normalized)
        normalized = parsed.netloc or parsed.path

    if normalized.startswith("www."):
        normalized = normalized[4:]

    normalized = normalized.split("/", 1)[0]
    normalized = normalized.split(":", 1)[0]
    if " " in normalized:
        normalized = normalized.split()[-1]
    if "." in normalized:
        normalized = normalized.split(".", 1)[0]

    aliases = {
        "ddl": "ddownload",
        "ddlto": "ddownload",
        "rg": "rapidgator",
    }
    return aliases.get(normalized, normalized)



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

from quasarr.constants import QUASARR_AVATAR, SPONSORS_HELPER_URL
from quasarr.providers.log import info
from quasarr.providers.notifications.helpers.common import (
    build_solved_data,
    format_balance,
    format_number,
)
from quasarr.providers.notifications.helpers.notification_message import (
    NotificationFact,
    NotificationFactsEntry,
    NotificationLinkEntry,
    NotificationMessage,
    NotificationValueEntry,
)
from quasarr.providers.notifications.helpers.notification_types import (
    NotificationType,
    normalize_notification_type,
)


def _build_solved_entries(details):
    data = build_solved_data(details)
    if data is None:
        return ()

    entries = []
    for solver in data.get("solvers", []):
        facts = []
        attempts = solver["attempts"] if solver.get("has_attempts") else 0
        facts.append(NotificationFact("Attempts", str(attempts)))

        currency = solver.get("currency")
        if solver.get("cost") is not None:
            cost_text = format_number(solver["cost"])
            if currency:
                cost_text = f"{cost_text} {currency}"
            facts.append(NotificationFact("Cost", cost_text))

        if solver.get("balance") is not None:
            balance_text = format_balance(solver["balance"])
            if currency:
                balance_text = f"{balance_text} {currency}"
            facts.append(NotificationFact("Balance", balance_text))

        if facts:
            entries.append(
                NotificationFactsEntry(
                    title=solver["solver_display"],
                    facts=tuple(facts),
                )
            )

    if data.get("duration"):
        entries.append(NotificationValueEntry(title="Duration", value=data["duration"]))

    return tuple(entries)


def _build_reason_entry(details):
    if not isinstance(details, dict):
        return None

    reason = details.get("reason") or details.get("error")
    if not reason:
        return None

    return NotificationValueEntry(title="Reason", value=str(reason))


def build_notification_message(
    shared_state,
    title,
    case,
    details=None,
    source=None,
    image_url=None,
):
    notification_type = normalize_notification_type(case)
    if notification_type is None:
        info(f"Unknown notification case: {case}")
        return None

    entries = []
    thumbnail_url = None

    if notification_type == NotificationType.UNPROTECTED:
        description = "No CAPTCHA required. Links were added directly!"
    elif notification_type == NotificationType.SOLVED:
        description = "CAPTCHA solved by SponsorsHelper!"
        entries.extend(_build_solved_entries(details))
    elif notification_type == NotificationType.FAILED:
        description = "Package marked as failed."
    elif notification_type == NotificationType.DISABLED:
        description = (
            "SponsorsHelper failed to solve the CAPTCHA! "
            "Please solve it manually to proceed."
        )
    elif notification_type == NotificationType.CAPTCHA:
        description = "Download will proceed, once the CAPTCHA has been solved."
        captcha_url = f"{shared_state.values['external_address']}/captcha"
        entries.append(
            NotificationLinkEntry(
                title="Solve CAPTCHA",
                text="Open this link to solve the CAPTCHA.",
                link_text="this link",
                url=captcha_url,
            )
        )
        if not shared_state.values.get("helper_active"):
            entries.append(
                NotificationLinkEntry(
                    title="SponsorsHelper",
                    text="Sponsors get automated CAPTCHA solutions!",
                    url=SPONSORS_HELPER_URL,
                )
            )
    elif notification_type == NotificationType.QUASARR_UPDATE:
        version = "latest"
        link = ""
        if isinstance(details, dict):
            version = details.get("version") or version
            link = details.get("link") or ""
        description = f"Please update to {version} as soon as possible!"
        if link:
            entries.append(
                NotificationLinkEntry(
                    title="Release notes at:",
                    text=f"GitHub.com: rix1337/Quasarr/{version}",
                    url=link,
                )
            )
        if not image_url:
            thumbnail_url = QUASARR_AVATAR
    elif notification_type == NotificationType.TEST:
        description = "This is a test notification from Quasarr UI configuration."
    else:
        info(f"Unknown notification case: {case}")
        return None

    if reason_entry := _build_reason_entry(details):
        entries.append(reason_entry)

    if source and source.startswith("http"):
        entries.append(
            NotificationLinkEntry(
                title="Source",
                text="View release details here",
                url=source,
            )
        )

    return NotificationMessage(
        title=title,
        description=description,
        entries=tuple(entries),
        image_url=image_url,
        thumbnail_url=thumbnail_url,
    )



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import html
import re
import time

from quasarr.constants import (
    FEED_REQUEST_TIMEOUT_SECONDS,
    SEARCH_CAT_MOVIES,
    SEARCH_CAT_SHOWS,
    SEARCH_REQUEST_TIMEOUT_SECONDS,
)
from quasarr.downloads.sources.rm import (
    _bootstrap_session,
    _build_release_page_url,
    _convert_iso_to_rss_date,
    _fetch_productions_page,
    _fetch_releases_by_production_ids,
    _get_base_url,
    _is_movie_release,
    _search_productions,
)
from quasarr.providers import shared_state
from quasarr.providers.hostname_issues import clear_hostname_issue, mark_hostname_issue
from quasarr.providers.imdb_metadata import get_localized_title, get_year
from quasarr.providers.log import debug, info, trace, warn
from quasarr.providers.utils import (
    generate_download_link,
    get_base_search_category_id,
    is_imdb_id,
    is_valid_release,
    release_matches_search_category,
    search_string_in_sanitized_title,
)
from quasarr.search.sources.helpers.search_release import SearchRelease
from quasarr.search.sources.helpers.search_source import AbstractSearchSource

_LEADING_TAG_REGEX = re.compile(r"^(?:\[[^\]]+\]\s*)+")


class Source(AbstractSearchSource):
    initials = "rm"
    supports_imdb = True
    supports_phrase = True
    supported_categories = [SEARCH_CAT_MOVIES, SEARCH_CAT_SHOWS]

    def feed(
        self, shared_state: shared_state, start_time: float, search_category: str
    ) -> list[SearchRelease]:
        releases = []

        if get_base_search_category_id(search_category) not in (
            SEARCH_CAT_MOVIES,
            SEARCH_CAT_SHOWS,
        ):
            return releases

        try:
            release_items = _load_feed_release_items(
                shared_state,
                search_category,
                max_production_pages=10,
                timeout=FEED_REQUEST_TIMEOUT_SECONDS,
            )
            releases = _build_search_results(
                shared_state,
                release_items,
                search_category,
                search_string="",
                imdb_id=None,
                is_feed=True,
            )
            releases = releases[:20]
        except Exception as e:
            warn(f"Error loading feed: {e}")
            mark_hostname_issue(
                self.initials,
                "feed",
                str(e) if "e" in dir() else "Error occurred",
            )

        elapsed_time = time.time() - start_time
        debug(f"Time taken: {elapsed_time:.2f}s")

        if releases:
            clear_hostname_issue(self.initials)
        return releases

    def search(
        self,
        shared_state: shared_state,
        start_time: float,
        search_category: str,
        search_string: str = "",
        season: int = None,
        episode: int = None,
    ) -> list[SearchRelease]:
        releases = []
        match_search_string = search_string

        if not search_string:
            return releases

        imdb_id = is_imdb_id(search_string)
        if imdb_id:
            localized_title = get_localized_title(shared_state, imdb_id, "en")
            if not localized_title:
                info(f"Could not extract title from IMDb-ID {imdb_id}")
                return releases
            search_string = html.unescape(localized_title)
            match_search_string = search_string
            if get_base_search_category_id(search_category) == SEARCH_CAT_MOVIES and (
                year := get_year(imdb_id)
            ):
                search_string += f" {year}"

        try:
            release_items = []
            search_session, productions = _load_search_productions(
                shared_state,
                search_string,
                SEARCH_REQUEST_TIMEOUT_SECONDS,
            )
            if productions:
                release_items = _fetch_releases_by_production_ids(
                    shared_state,
                    [production.get("id") for production in productions],
                    SEARCH_REQUEST_TIMEOUT_SECONDS,
                    session=search_session,
                )

            releases = _build_search_results(
                shared_state,
                release_items,
                search_category,
                search_string=match_search_string,
                imdb_id=imdb_id,
                season=season,
                episode=episode,
            )
        except Exception as e:
            warn(f"Error loading search: {e}")
            mark_hostname_issue(
                self.initials,
                "search",
                str(e) if "e" in dir() else "Error occurred",
            )

        elapsed_time = time.time() - start_time
        debug(f"Time taken: {elapsed_time:.2f}s")

        if releases:
            clear_hostname_issue(self.initials)
        return releases


def _load_feed_release_items(
    shared_state, search_category, max_production_pages, timeout
):
    session = _bootstrap_session(shared_state, timeout)
    production_ids = []
    base_search_category = get_base_search_category_id(search_category)

    for page in range(1, max_production_pages + 1):
        payload = _fetch_productions_page(
            shared_state,
            page,
            timeout,
            session=session,
        )
        productions = payload.get("productions") or []
        if not productions:
            break

        for production in productions:
            if not _matches_production_category(production, base_search_category):
                continue
            production_id = production.get("id")
            if production_id:
                production_ids.append(production_id)

    if not production_ids:
        return []

    return _fetch_releases_by_production_ids(
        shared_state,
        production_ids,
        timeout,
        session=session,
    )


def _build_search_variants(search_string):
    normalized = " ".join(str(search_string or "").split())
    if not normalized:
        return []

    normalized = re.sub(r"[^\w\s]+", " ", normalized)
    normalized = " ".join(normalized.split())
    words = normalized.split()
    variants = []

    def append_word_variants(parts):
        if not parts:
            return

        variants.append(" ".join(parts))

        if len(parts) > 3:
            variants.append(" ".join(parts[:2]))
            variants.append(" ".join(parts[-3:]))
        elif len(parts) == 3:
            variants.append(parts[0])
            variants.append(" ".join(parts[-2:]))
        elif len(parts) == 2:
            variants.append(parts[0])
            variants.append(parts[1])

    append_word_variants(words)

    if words and words[-1].isdigit() and len(words[-1]) == 4:
        append_word_variants(words[:-1])

    seen = set()
    ordered = []
    for variant in variants:
        cleaned = " ".join(variant.split()).strip()
        if not cleaned or cleaned in seen:
            continue
        seen.add(cleaned)
        ordered.append(cleaned)

    return ordered


def _load_search_productions(shared_state, search_string, timeout):
    session = _bootstrap_session(shared_state, timeout)
    productions_by_id = {}

    for variant in _build_search_variants(search_string):
        try:
            productions = _search_productions(
                shared_state,
                variant,
                timeout,
                session=session,
            )
        except Exception as e:
            trace(f'RM production search failed for variant "{variant}": {e}')
            continue
        for production in productions:
            production_id = production.get("id")
            if production_id and production_id not in productions_by_id:
                productions_by_id[production_id] = production

    return session, list(productions_by_id.values())


def _matches_production_category(production, base_search_category):
    medium_slug = (
        str((production.get("medium") or {}).get("slug") or "").strip().lower()
    )
    if not medium_slug:
        return False

    if base_search_category == SEARCH_CAT_MOVIES:
        return medium_slug == "movie"

    if base_search_category == SEARCH_CAT_SHOWS:
        return medium_slug != "movie"

    return False


def _matches_release_category(search_category, release):
    base_search_category = get_base_search_category_id(search_category)
    is_movie_release = _is_movie_release(release)

    if base_search_category == SEARCH_CAT_MOVIES:
        return is_movie_release

    if base_search_category == SEARCH_CAT_SHOWS:
        return not is_movie_release

    return False


def _matches_requested_release(
    release,
    title,
    search_category,
    search_string,
    season=None,
    episode=None,
):
    base_search_category = get_base_search_category_id(search_category)
    if base_search_category != SEARCH_CAT_SHOWS:
        return is_valid_release(title, search_category, search_string, season, episode)

    release_season = _normalize_release_number(release.get("season"))
    release_episode = _normalize_release_number(release.get("episode"))
    if release_season is None and release_episode is None:
        return is_valid_release(title, search_category, search_string, season, episode)

    if not search_string_in_sanitized_title(search_string, title):
        return False

    if season is not None:
        if release_season is None:
            return is_valid_release(
                title, search_category, search_string, season, episode
            )
        if release_season != int(season):
            return False

    if episode is not None:
        if release_episode is None:
            return is_valid_release(
                title, search_category, search_string, season, episode
            )
        if release_episode != int(episode):
            return False

    return True


def _normalize_release_number(value):
    try:
        return int(value)
    except (TypeError, ValueError):
        return None


def _normalize_display_title(title):
    normalized = html.unescape(str(title or "").strip())
    normalized = _LEADING_TAG_REGEX.sub("", normalized)
    normalized = " ".join(normalized.split())
    return normalized.replace(" ", ".")


def _build_search_results(
    shared_state,
    release_items,
    search_category,
    search_string,
    imdb_id,
    season=None,
    episode=None,
    is_feed=False,
):
    base_url = _get_base_url(shared_state)
    releases = []
    seen_sources = set()

    for release in release_items:
        try:
            if not _matches_release_category(search_category, release):
                continue

            raw_title = html.unescape(str(release.get("title") or "").strip())
            if not raw_title:
                continue

            if not release_matches_search_category(search_category, raw_title):
                continue

            if not is_feed and not _matches_requested_release(
                release,
                raw_title,
                search_category,
                search_string,
                season,
                episode,
            ):
                continue

            title = _normalize_display_title(raw_title)

            slug = str(release.get("slug") or "").strip()
            if not slug:
                continue

            source = _build_release_page_url(base_url, slug)
            if source in seen_sources:
                continue
            seen_sources.add(source)

            size_bytes = int(release.get("size") or 0)
            size_mb = size_bytes / (1024 * 1024) if size_bytes else 0
            published = _convert_iso_to_rss_date(
                release.get("updated_at") or release.get("created_at")
            )
            release_imdb_id = imdb_id

            releases.append(
                {
                    "details": {
                        "title": title,
                        "hostname": Source.initials,
                        "imdb_id": release_imdb_id,
                        "link": generate_download_link(
                            shared_state,
                            title,
                            source,
                            size_mb,
                            "",
                            release_imdb_id,
                            Source.initials,
                        ),
                        "size": size_bytes,
                        "date": published,
                        "source": source,
                    },
                    "type": "protected",
                }
            )
        except Exception as e:
            warn(f"Error parsing release: {e}")
            continue

    return releases



# -*- coding: utf-8 -*-

import re
from urllib.parse import urlparse

from quasarr.constants import (
    MIRROR_TOKEN_ALIASES,
    SHARE_HOSTERS_LOWERCASE,
)

_MULTI_LABEL_SUFFIX_PREFIXES = frozenset(
    {"ac", "co", "com", "edu", "gov", "net", "org"}
)
_TOKEN_CLEAN_RE = re.compile(r"[^a-z0-9]+")


def _clean_token(value):
    return _TOKEN_CLEAN_RE.sub("", str(value or "").lower())


_CANONICAL_MIRROR_TOKENS = {_clean_token(hoster) for hoster in SHARE_HOSTERS_LOWERCASE}
_MIRROR_TOKEN_MAP = {
    _clean_token(alias): _clean_token(canonical)
    for alias, canonical in MIRROR_TOKEN_ALIASES.items()
}
for canonical_token in _CANONICAL_MIRROR_TOKENS:
    _MIRROR_TOKEN_MAP[canonical_token] = canonical_token


def normalize_mirror_token(value):
    """
    Normalize a mirror name, hostname, or URL to the canonical hoster token.

    Matching intentionally ignores TLDs and subdomains and focuses on the
    provider/root label of the final HTTP hostname.
    """
    host_or_name = _extract_host_or_name(value)
    if not host_or_name:
        return ""

    if "." in host_or_name:
        return _normalize_root_token(_extract_domain_root_token(host_or_name))

    return _normalize_root_token(host_or_name)


def filter_final_download_urls(urls, mirrors):
    allowed_tokens = {
        normalize_mirror_token(mirror) for mirror in (mirrors or []) if mirror
    }
    allowed_tokens.discard("")

    normalized_urls = [str(url) for url in (urls or []) if url]
    if not allowed_tokens:
        return {
            "urls": normalized_urls,
            "allowed_tokens": allowed_tokens,
            "dropped": [],
            "kept_tokens": {
                normalize_mirror_token(url) for url in normalized_urls if url
            },
        }

    kept_urls = []
    dropped = []
    kept_tokens = set()

    for url in normalized_urls:
        token = normalize_mirror_token(url)
        if token and token in allowed_tokens:
            kept_urls.append(url)
            kept_tokens.add(token)
            continue
        dropped.append({"url": url, "token": token or "unknown"})

    return {
        "urls": kept_urls,
        "allowed_tokens": allowed_tokens,
        "dropped": dropped,
        "kept_tokens": kept_tokens,
    }


def _extract_host_or_name(value):
    raw = str(value or "").strip().lower()
    if not raw:
        return ""

    if "://" in raw or raw.startswith("//"):
        parsed = urlparse(raw if "://" in raw else f"https:{raw}")
        return (parsed.hostname or "").lower().strip(".")

    raw = raw.split("/", 1)[0]
    raw = raw.rsplit("@", 1)[-1]
    raw = raw.split(":", 1)[0]
    return raw.lower().strip(".")


def _extract_domain_root_token(hostname):
    labels = [label for label in hostname.split(".") if label]
    if not labels:
        return ""
    if len(labels) == 1:
        return labels[0]
    if (
        len(labels) >= 3
        and len(labels[-1]) == 2
        and _clean_token(labels[-2]) in _MULTI_LABEL_SUFFIX_PREFIXES
    ):
        return labels[-3]
    return labels[-2]


def _normalize_root_token(token):
    cleaned = _clean_token(token)
    if not cleaned:
        return ""
    return _MIRROR_TOKEN_MAP.get(cleaned, cleaned)



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

from quasarr.providers.notifications.helpers.abstract_notification_formatter import (
    AbstractNotificationFormatter,
)
from quasarr.providers.notifications.helpers.common import (
    build_solved_data,
    canonicalize_solver_name,
    format_balance,
    format_duration,
    format_number,
    resolve_poster_url,
)
from quasarr.providers.notifications.helpers.message_builder import (
    build_notification_message,
)
from quasarr.providers.notifications.helpers.notification_message import (
    AbstractNotificationEntry,
    NotificationFact,
    NotificationFactsEntry,
    NotificationLinkEntry,
    NotificationMessage,
    NotificationTextEntry,
    NotificationValueEntry,
)
from quasarr.providers.notifications.helpers.notification_types import (
    NotificationType,
    get_notification_type_label,
    get_user_configurable_notification_types,
    normalize_notification_type,
)

__all__ = [
    "AbstractNotificationEntry",
    "AbstractNotificationFormatter",
    "NotificationFact",
    "NotificationFactsEntry",
    "NotificationLinkEntry",
    "NotificationMessage",
    "NotificationTextEntry",
    "NotificationType",
    "NotificationValueEntry",
    "build_notification_message",
    "build_solved_data",
    "canonicalize_solver_name",
    "format_balance",
    "format_duration",
    "format_number",
    "get_notification_type_label",
    "get_user_configurable_notification_types",
    "normalize_notification_type",
    "resolve_poster_url",
]



# -*- coding: utf-8 -*-

import unittest
from types import SimpleNamespace

from quasarr.providers.notifications.helpers.message_builder import (
    build_notification_message,
)
from quasarr.providers.notifications.helpers.notification_types import NotificationType


class NotificationMessageBuilderTests(unittest.TestCase):
    def setUp(self):
        self.shared_state = SimpleNamespace(values={})

    def test_failed_notification_includes_reason_entry(self):
        message = build_notification_message(
            self.shared_state,
            "Example.Release",
            NotificationType.FAILED,
            details={"reason": "All final download links were rejected."},
        )

        self.assertIsNotNone(message)
        self.assertEqual("Package marked as failed.", message.description)
        self.assertEqual(1, len(message.entries))
        self.assertEqual("Reason", message.entries[0].title)
        self.assertEqual(
            "All final download links were rejected.",
            message.entries[0].value,
        )

    def test_disabled_notification_uses_error_as_reason(self):
        message = build_notification_message(
            self.shared_state,
            "Example.Release",
            NotificationType.DISABLED,
            details={"error": "SponsorsHelper hit its retry limit."},
        )

        self.assertIsNotNone(message)
        self.assertEqual(1, len(message.entries))
        self.assertEqual("Reason", message.entries[0].title)
        self.assertEqual(
            "SponsorsHelper hit its retry limit.",
            message.entries[0].value,
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import base64
import hashlib
import hmac
import html
import json
import os
import time
from functools import wraps
from urllib.parse import urlsplit

from bottle import abort, redirect, request, response

import quasarr.providers.html_images as images
from quasarr.providers.version import get_version
from quasarr.storage.config import Config

# Auth configuration from environment
_AUTH_USER = os.environ.get("USER", "")
_AUTH_PASS = os.environ.get("PASS", "")
_AUTH_TYPE = os.environ.get("AUTH", "form").lower()

# Cookie settings
_COOKIE_NAME = "quasarr_session"
_COOKIE_MAX_AGE = 30 * 24 * 60 * 60  # 30 days

# Stable secret derived from PASS (restart-safe)
_SECRET_KEY = hashlib.sha256(_AUTH_PASS.encode("utf-8")).digest()

_AUTH_MODE_ATTR = "__quasarr_auth_mode__"
_AUTH_MODE_PUBLIC = "public"
_AUTH_MODE_BROWSER = "browser"
_AUTH_MODE_API_KEY = "api_key"


def is_auth_enabled():
    """Check if authentication is enabled (both USER and PASS set)."""
    return bool(_AUTH_USER and _AUTH_PASS)


def is_form_auth():
    """Check if form-based auth is enabled."""
    return _AUTH_TYPE == "form"


def _b64encode(data: bytes) -> str:
    return base64.urlsafe_b64encode(data).decode("ascii").rstrip("=")


def _b64decode(data: str) -> bytes:
    padding = "=" * (-len(data) % 4)
    return base64.urlsafe_b64decode(data + padding)


def _sign(data: bytes) -> bytes:
    return hmac.new(_SECRET_KEY, data, hashlib.sha256).digest()


def _mask_user(user: str) -> str:
    """
    One-way masked user identifier.
    Stable across restarts, not reversible.
    """
    return hashlib.sha256(f"user:{user}".encode("utf-8")).hexdigest()


def _create_session_cookie(user: str) -> str:
    """
    Stateless, signed cookie.
    Stores only masked user + expiry.
    """
    payload = {
        "u": _mask_user(user),
        "exp": int(time.time()) + _COOKIE_MAX_AGE,
    }
    raw = json.dumps(payload, separators=(",", ":")).encode("utf-8")
    sig = _sign(raw)
    return f"{_b64encode(raw)}.{_b64encode(sig)}"


def _invalidate_cookie():
    try:
        response.delete_cookie(_COOKIE_NAME, path="/")
    except Exception:
        pass


def _verify_session_cookie(value: str) -> bool:
    """
    Verify signature, expiry, and masked user.
    On ANY failure → force logout (cookie deletion).
    """
    try:
        if not value or "." not in value:
            raise ValueError

        raw_b64, sig_b64 = value.split(".", 1)
        raw = _b64decode(raw_b64)
        sig = _b64decode(sig_b64)

        if not hmac.compare_digest(sig, _sign(raw)):
            raise ValueError

        payload = json.loads(raw.decode("utf-8"))

        if payload.get("u") != _mask_user(_AUTH_USER):
            raise ValueError

        if int(time.time()) > int(payload.get("exp", 0)):
            raise ValueError

        return True
    except Exception:
        _invalidate_cookie()
        return False


def check_basic_auth():
    """Check HTTP Basic Auth header. Returns True if valid."""
    auth = request.headers.get("Authorization", "")
    if not auth.startswith("Basic "):
        return False
    try:
        decoded = base64.b64decode(auth[6:]).decode("utf-8")
        user, passwd = decoded.split(":", 1)
        return user == _AUTH_USER and passwd == _AUTH_PASS
    except:
        return False


def check_form_auth():
    """Check session cookie. Returns True if valid."""
    cookie = request.get_cookie(_COOKIE_NAME)
    return bool(cookie and _verify_session_cookie(cookie))


def require_basic_auth():
    """Send 401 response for Basic Auth."""
    response.status = 401
    response.set_header("WWW-Authenticate", 'Basic realm="Quasarr"')
    return "Authentication required"


def is_browser_authenticated():
    """Check whether the current browser request is authenticated."""
    if not is_auth_enabled():
        return False
    if is_form_auth():
        return check_form_auth()
    return check_basic_auth()


def _mark_auth_mode(func, mode):
    setattr(func, _AUTH_MODE_ATTR, mode)
    return func


def public_endpoint(func):
    """Mark a route as public."""
    return _mark_auth_mode(func, _AUTH_MODE_PUBLIC)


def _normalize_next_url(next_url):
    """Allow only internal post-login redirects."""
    if not next_url:
        return "/"

    candidate = str(next_url).strip()
    if not candidate.startswith("/") or candidate.startswith("//"):
        return "/"

    parsed = urlsplit(candidate)
    if parsed.scheme or parsed.netloc or not parsed.path.startswith("/"):
        return "/"
    if parsed.path in {"/login", "/logout"}:
        return "/"

    normalized = parsed.path or "/"
    if parsed.query:
        normalized += f"?{parsed.query}"
    return normalized


def _render_login_page(error=None):
    """Render login form page using Quasarr styling."""
    error_html = (
        f'<p style="color: #dc3545; margin-bottom: 1rem;"><b>{html.escape(str(error))}</b></p>'
        if error
        else ""
    )
    next_url = html.escape(
        _normalize_next_url(request.query.get("next", "/")), quote=True
    )

    # Inline the centered HTML to avoid circular import
    return f'''<html>
    <head>
        <meta charset="utf-8">
        <meta name="viewport" content="width=device-width, initial-scale=1">
        <title>Quasarr - Login</title>
        <link rel="icon" href="{images.favicon}" type="image/png" sizes="16x16">
        <style>
            :root {{
                --bg-color: #ffffff;
                --fg-color: #212529;
                --card-bg: #ffffff;
                --card-shadow: rgba(0, 0, 0, 0.1);
                --primary: #0d6efd;
                --secondary: #6c757d;
                --spacing: 1rem;
            }}
            @media (prefers-color-scheme: dark) {{
                :root {{
                    --bg-color: #181a1b;
                    --fg-color: #f1f1f1;
                    --card-bg: #242526;
                    --card-shadow: rgba(0, 0, 0, 0.5);
                }}
            }}
            *, *::before, *::after {{ box-sizing: border-box; }}
            html, body {{
                margin: 0; padding: 0; width: 100%; height: 100%;
                background-color: var(--bg-color); color: var(--fg-color);
                font-family: system-ui, -apple-system, 'Segoe UI', Roboto, sans-serif;
                line-height: 1.6; display: flex; flex-direction: column; min-height: 100vh;
            }}
            .outer {{ flex: 1; display: flex; justify-content: center; align-items: center; padding: var(--spacing); }}
            .inner {{
                background-color: var(--card-bg); border-radius: 1rem;
                box-shadow: 0 0.5rem 1.5rem var(--card-shadow);
                padding: calc(var(--spacing) * 2); text-align: center; width: 100%; max-width: 400px;
            }}
            .logo {{ height: 64px; width: 64px; vertical-align: middle; margin-right: 0.5rem; }}
            h1 {{ margin: 0 0 1rem 0; font-size: 1.75rem; }}
            h2 {{ margin: 0 0 1.5rem 0; font-size: 1.25rem; font-weight: 500; }}
            label {{ display: block; text-align: left; margin-bottom: 0.25rem; font-weight: 500; }}
            input[type="text"], input[type="password"] {{
                width: 100%; padding: 0.5rem; margin-bottom: 1rem;
                border: 1px solid var(--secondary); border-radius: 0.5rem;
                font-size: 1rem; background: var(--card-bg); color: var(--fg-color);
            }}
            .btn-primary {{
                background-color: var(--primary); color: #fff; border: 1.5px solid #0856c7;
                padding: 0.5rem 1.5rem; font-size: 1rem; border-radius: 0.5rem;
                font-weight: 500; cursor: pointer; width: 100%;
            }}
            .btn-primary:hover {{ background-color: #0b5ed7; }}
            footer {{ text-align: center; font-size: 0.75rem; color: var(--secondary); padding: 0.5rem 0; }}
        </style>
    </head>
    <body>
        <div class="outer">
            <div class="inner">
                <h1><img src="{images.logo}" class="logo" alt="Quasarr"/>Quasarr</h1>
                {error_html}
                <form method="post" action="/login">
                    <input type="hidden" name="next" value="{next_url}">
                    <label for="username">Username</label>
                    <input type="text" id="username" name="username" autocomplete="username" required>
                    <label for="password">Password</label>
                    <input type="password" id="password" name="password" autocomplete="current-password" required>
                    <button type="submit" class="btn-primary">Login</button>
                </form>
            </div>
        </div>
        <footer>Quasarr v.{get_version()}</footer>
    </body>
    </html>'''


def _handle_login_post():
    """Handle login form submission."""
    username = request.forms.get("username", "")
    password = request.forms.get("password", "")
    next_url = _normalize_next_url(request.forms.get("next", "/"))

    if username == _AUTH_USER and password == _AUTH_PASS:
        cookie = _create_session_cookie(username)
        secure_flag = request.url.startswith("https://")
        response.set_cookie(
            _COOKIE_NAME,
            cookie,
            max_age=_COOKIE_MAX_AGE,
            path="/",
            httponly=True,
            secure=secure_flag,
            samesite="Lax",
        )
        redirect(next_url)
    else:
        _invalidate_cookie()
        return _render_login_page("Invalid username or password")


def _handle_logout():
    _invalidate_cookie()
    redirect("/login")


def show_logout_link():
    """Returns True if logout link should be shown (form auth enabled and authenticated)."""
    return is_auth_enabled() and is_form_auth() and check_form_auth()


def add_auth_routes(app):
    """Add login/logout routes to a Bottle app (for form auth only)."""
    if not is_auth_enabled():
        return

    if is_form_auth():

        @app.get("/login")
        @public_endpoint
        def login_get():
            if check_form_auth():
                redirect("/")
            return _render_login_page()

        @app.post("/login")
        @public_endpoint
        def login_post():
            return _handle_login_post()

        @app.get("/logout")
        def logout():
            return _handle_logout()


def add_auth_hook(app, whitelist=None):
    """Install the Quasarr authentication policy on a Bottle app.

    Args:
        app: Bottle application
        whitelist: List of path prefixes or suffixes to keep public
    """
    if whitelist is None:
        whitelist = []

    if any(getattr(plugin, "name", None) == "quasarr-auth" for plugin in app.plugins):
        return

    class AuthPlugin:
        name = "quasarr-auth"
        api = 2

        def __init__(self, public_whitelist):
            self.public_whitelist = tuple(public_whitelist)

        def _is_public_rule(self, rule):
            for item in self.public_whitelist:
                if rule.startswith(item) or rule.endswith(item):
                    return True
            return False

        def _get_auth_mode(self, callback, route):
            if self._is_public_rule(route.rule):
                return _AUTH_MODE_PUBLIC
            return getattr(callback, _AUTH_MODE_ATTR, _AUTH_MODE_BROWSER)

        def apply(self, callback, route):
            mode = self._get_auth_mode(callback, route)
            if mode == _AUTH_MODE_PUBLIC:
                return callback
            if mode == _AUTH_MODE_API_KEY:

                @wraps(callback)
                def api_key_wrapper(*args, **kwargs):
                    api_key = Config("API").get("key")
                    request_api_key = _get_request_api_key()
                    if not request_api_key:
                        return abort(401, "Missing API Key")
                    if request_api_key != api_key:
                        return abort(403, "Invalid API Key")
                    return callback(*args, **kwargs)

                return api_key_wrapper

            @wraps(callback)
            def browser_wrapper(*args, **kwargs):
                if not is_auth_enabled():
                    return callback(*args, **kwargs)

                path = request.path.split("?")[0]

                if is_form_auth():
                    if not check_form_auth():
                        _invalidate_cookie()
                        if path == "/logout":
                            redirect("/login")
                        redirect(f"/login?next={path}")
                else:
                    if not check_basic_auth():
                        return require_basic_auth()

                return callback(*args, **kwargs)

            return browser_wrapper

    app.install(AuthPlugin(whitelist))


def _get_request_api_key():
    api_key = request.query.get("apikey") or request.forms.get("apikey")
    if api_key:
        return api_key

    api_key = request.headers.get("X-API-Key") or request.headers.get("X-Api-Key")
    if api_key:
        return api_key

    auth_header = request.headers.get("Authorization", "")
    if auth_header.startswith("Bearer "):
        return auth_header[7:].strip()

    return None


def require_api_key(func):
    """Mark a route to require the Quasarr API key."""
    return _mark_auth_mode(func, _AUTH_MODE_API_KEY)


def require_browser_auth(func):
    """Mark a route to require browser authentication."""
    return _mark_auth_mode(func, _AUTH_MODE_BROWSER)


def audit_route_auth_modes(app, api_key_prefixes=None, public_whitelist=None):
    """Fail fast if protected route groups are not explicitly API-key guarded."""
    if api_key_prefixes is None:
        api_key_prefixes = ()
    if public_whitelist is None:
        public_whitelist = ()

    violations = []
    for route in app.routes:
        rule = route.rule

        is_public = any(
            rule.startswith(item) or rule.endswith(item) for item in public_whitelist
        )
        mode = (
            _AUTH_MODE_PUBLIC
            if is_public
            else getattr(route.callback, _AUTH_MODE_ATTR, _AUTH_MODE_BROWSER)
        )

        if any(rule.startswith(prefix) for prefix in api_key_prefixes):
            if mode != _AUTH_MODE_API_KEY:
                violations.append(f"{route.method} {rule} -> {mode}")

    if violations:
        joined = ", ".join(sorted(violations))
        raise RuntimeError(f"Auth policy violation: {joined}")



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import json

import requests

from quasarr.constants import (
    QUASARR_AVATAR,
    SESSION_REQUEST_TIMEOUT_SECONDS,
    SUPPRESS_NOTIFICATIONS,
)
from quasarr.providers.log import info
from quasarr.providers.notifications.helpers.abstract_notification_formatter import (
    AbstractNotificationFormatter,
)
from quasarr.providers.notifications.helpers.notification_message import (
    NotificationFactsEntry,
    NotificationLinkEntry,
    NotificationMessage,
    NotificationTextEntry,
    NotificationValueEntry,
)


class DiscordNotificationFormatter(AbstractNotificationFormatter):
    @staticmethod
    def _format_link(text, url, link_text=None):
        target_text = link_text or text
        if target_text and target_text in text:
            return text.replace(target_text, f"[{target_text}]({url})", 1)
        return f"[{text}]({url})"

    def render_text_entry(self, entry: NotificationTextEntry):
        return {"name": entry.title, "value": entry.text}

    def render_link_entry(self, entry: NotificationLinkEntry):
        return {
            "name": entry.title,
            "value": self._format_link(entry.text, entry.url, entry.link_text),
        }

    def render_facts_entry(self, entry: NotificationFactsEntry):
        return {
            "name": entry.title,
            "value": " | ".join(
                f"**{fact.label}:** {fact.value}" for fact in entry.facts
            ),
        }

    def render_value_entry(self, entry: NotificationValueEntry):
        return {"name": entry.title, "value": entry.value}

    def render_message(self, message: NotificationMessage):
        embed = {"title": message.title, "description": message.description}
        fields = self.render_entries(message.entries)
        if fields:
            embed["fields"] = fields

        if message.image_url:
            poster_object = {"url": message.image_url}
            embed["thumbnail"] = poster_object
            embed["image"] = poster_object
        elif message.thumbnail_url:
            embed["thumbnail"] = {"url": message.thumbnail_url}

        return embed


def _get_discord_webhook(shared_state):
    settings = shared_state.values.get("notification_settings")
    if not isinstance(settings, dict):
        return ""
    return str(settings.get("discord_webhook") or "").strip()


def send(shared_state, message, silent=True):
    """Send a rendered Discord webhook notification. Returns True on success."""
    webhook_url = _get_discord_webhook(shared_state)
    if not webhook_url:
        return False

    if not isinstance(message, NotificationMessage):
        info(f"Invalid Discord notification payload: {type(message).__name__}")
        return False

    embed = DiscordNotificationFormatter().render_message(message)
    data = {
        "username": "Quasarr",
        "avatar_url": QUASARR_AVATAR,
        "embeds": [embed],
    }

    if silent:
        data["flags"] = SUPPRESS_NOTIFICATIONS

    response = requests.post(
        webhook_url,
        data=json.dumps(data),
        headers={"Content-Type": "application/json"},
        timeout=SESSION_REQUEST_TIMEOUT_SECONDS,
    )
    if response.status_code != 204:
        info(
            f"Failed to send message to Discord webhook. "
            f"Status code: {response.status_code}"
        )
        return False
    return True



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import os
import sys

from bottle import Bottle, request

import quasarr.providers.web_server
from quasarr.providers.html_templates import render_button, render_form
from quasarr.providers.log import info
from quasarr.providers.web_server import Server
from quasarr.storage.setup.common import (
    add_no_cache_headers,
    render_reconnect_success,
    setup_auth,
)


def path_config(shared_state):
    app = Bottle()
    add_no_cache_headers(app)
    setup_auth(app)

    current_path = os.path.dirname(os.path.abspath(sys.argv[0]))

    @app.get("/")
    def config_form():
        config_form_html = f'''
            <form action="/api/config" method="post" onsubmit="return handleSubmit(this)">
                <label for="config_path">Path</label>
                <input type="text" id="config_path" name="config_path" placeholder="{current_path}"><br>
                {render_button("Save", "primary", {"type": "submit", "id": "submitBtn"})}
            </form>
            <script>
            var formSubmitted = false;
            function handleSubmit(form) {{
                if (formSubmitted) return false;
                formSubmitted = true;
                var btn = document.getElementById('submitBtn');
                if (btn) {{ btn.disabled = true; btn.textContent = 'Saving...'; }}
                return true;
            }}
            </script>
            '''
        return render_form(
            "Press 'Save' to set desired path for configuration", config_form_html
        )

    def set_config_path(config_path):
        config_path_file = "Quasarr.conf"

        if not config_path:
            config_path = current_path

        config_path = config_path.replace("\\", "/")
        config_path = config_path[:-1] if config_path.endswith("/") else config_path

        if not os.path.exists(config_path):
            os.makedirs(config_path)

        with open(config_path_file, "w") as f:
            f.write(config_path)

        return config_path

    @app.post("/api/config")
    def set_config():
        config_path = request.forms.get("config_path")
        config_path = set_config_path(config_path)
        quasarr.providers.web_server.temp_server_success = True
        return render_reconnect_success(f'Config path set to: "{config_path}"')

    info(
        f'Starting web server for config at: "{shared_state.values["external_address"]}".'
    )
    info("Please set desired config path there!")
    quasarr.providers.web_server.temp_server_success = False
    return Server(
        app, listen="0.0.0.0", port=shared_state.values["port"]
    ).serve_temporarily()
